import os
import logging
import pandas as pd
from dotenv import load_dotenv
import feedparser
import requests
import zipfile
import xml.etree.ElementTree as ET
from concurrent.futures import ThreadPoolExecutor, as_completed
import schedule
import time
from datetime import datetime

# Load environment variables from .env file located in the config directory
config_dir = 'C:\\Users\\user\\projects\\st\\config'
dotenv_path = os.path.join(config_dir, '.env')
load_dotenv(dotenv_path)

# API Key and URLs
dart_api_key = os.getenv('DART_API_KEY', 'd302fb1ed6c724285e694609359e277bfeed2a86')
ecos_api_key = os.getenv('ECOS_API_KEY', "JIUN3YBFR9JWR4Y9G2KT")
corp_code_url = 'https://opendart.fss.or.kr/api/corpCode.xml'
fnltt_url = 'https://opendart.fss.or.kr/api/fnlttSinglAcntAll.json'

# Settings
settings = {
    'data_path': r'C:\Users\user\projects\st\data',
    'db_path': r'C:\Users\user\projects\st\data\data.db',
    'stock_codes_path': r'C:\Users\user\projects\st\data\stock_codes.csv',
    'stock_n_codes_path': r'C:\Users\user\projects\st\data\stock_n_codes.csv',  # New path for stock_n_codes.csv
    'last_run_file': r'C:\Users\user\projects\st\data\last_run.txt',
    'financial_data_path': r'C:\Users\user\projects\st\data\financial_data',
    'save_path': r'C:\Users\user\projects\st\data\news_data',  # Path for saving news data
    'economic_data_path': r'C:\Users\user\projects\st\data\economic_data'  # Path for economic data
}

os.makedirs(settings['save_path'], exist_ok=True)
os.makedirs(settings['financial_data_path'], exist_ok=True)
os.makedirs(settings['economic_data_path'], exist_ok=True)

# Set logging level to INFO to disable debug logs
logging.basicConfig(level=logging.INFO)

# Indicators
indicators = {
    "GDP": "200Y001",  # 국내총생산 (GDP)
    "Inflation Rate": "901Y010"  # 소비자물가지수 (CPI)
}

# Function to get stock codes and names
def get_stock_codes_and_names():
    logging.info("Fetching stock codes and names...")
    file_path = settings['stock_codes_path']
    if not os.path.exists(file_path):
        logging.error(f"{file_path} not found.")
        return []
    df = pd.read_csv(file_path, dtype={'code': str})  # Ensure 'code' column is read as string
    return df.to_dict('records')

# Function to fetch and save the current stock codes to a file
def fetch_and_save_stock_codes():
    logging.info("Fetching current stock codes from source...")
    new_stock_codes = get_stock_codes_and_names()
    new_codes_df = pd.DataFrame(new_stock_codes)
    new_codes_df.to_csv(settings['stock_codes_path'], index=False)
    return new_stock_codes

# Function to compare and find new stock codes
def find_new_stock_codes(old_codes, new_codes):
    old_code_set = set([stock['code'] for stock in old_codes])
    new_code_set = set([stock['code'] for stock in new_codes])
    new_stock_codes = new_code_set - old_code_set
    return [code for code in new_codes if code['code'] in new_stock_codes]

# Function to get the last run date from a file
def get_last_run_date():
    if os.path.exists(settings['last_run_file']):
        with open(settings['last_run_file'], 'r') as f:
            return datetime.strptime(f.read().strip(), '%Y-%m-%d')
    return None

# Function to save the last run date to a file
def save_last_run_date():
    with open(settings['last_run_file'], 'w') as f:
        f.write(datetime.now().strftime('%Y-%m-%d'))

# Function to parse RSS feed
def parse_rss_feed(feed_url, stock_name, stock_code):
    logging.info(f"Parsing RSS feed for {stock_name} ({stock_code})...")
    news_items = []
    feed = feedparser.parse(feed_url)
    for entry in feed.entries:
        try:
            title = entry.title
            link = entry.link
            # Parse the date and format it as 'YYYY-MM-DD'
            date = datetime.strptime(entry.published, '%a, %d %b %Y %H:%M:%S %Z').strftime('%Y-%m-%d')
            news_items.append({'title': title, 'link': link, 'date': date, 'stock_name': stock_name, 'stock_code': stock_code})
        except Exception as e:
            logging.error(f"Error parsing RSS entry: {e}")
    return news_items

# Function to fetch news from Google
def fetch_news_from_google(stock_name, stock_code):
    logging.info(f"Fetching news from Google for {stock_name} ({stock_code})...")
    
    # 주식 이름에서 공백을 %20으로 인코딩 (stock_name을 문자열로 변환)
    if pd.isna(stock_name):  # stock_name이 NaN이면 빈 문자열로 처리
        encoded_stock_name = ""
    else:
        encoded_stock_name = str(stock_name).replace(" ", "%20")
    
    google_news_rss_url = f'https://news.google.com/rss/search?q={encoded_stock_name}&hl=ko&gl=KR&ceid=KR:ko'
    return parse_rss_feed(google_news_rss_url, stock_name, stock_code.zfill(6))

# Function to save news data
def save_news(news_items, save_path):
    if os.path.exists(save_path):
        try:
            existing_news_data = pd.read_csv(save_path, dtype={'stock_code': str})
            if existing_news_data.empty:
                existing_news_data = pd.DataFrame()
        except pd.errors.EmptyDataError:
            existing_news_data = pd.DataFrame()
        except Exception as e:
            logging.error(f"Error reading file {save_path}: {e}")
            return
    else:
        existing_news_data = pd.DataFrame()

    existing_links = set(existing_news_data['link']) if not existing_news_data.empty else set()
    new_news_data = [item for item in news_items if item['link'] not in existing_links]

    if not new_news_data:
        logging.info(f"No new news data to add for {os.path.basename(save_path).replace('_news_data.csv', '')}.")
        return

    df = pd.DataFrame(new_news_data)
    combined_data = pd.concat([existing_news_data, df], ignore_index=True)
    combined_data.to_csv(save_path, index=False)
    logging.info(f"News data saved to {save_path}")

# Function to update news data
def update_news_data(stock_code, stock_name, last_run_date):
    logging.info(f"Updating news data for {stock_name} ({stock_code})...")
    stock_code_padded = stock_code.zfill(6)  # 6자리로 맞추기
    all_news_data = fetch_news_from_google(stock_name, stock_code_padded)
    
    # Filter news by last run date
    if last_run_date:
        all_news_data = [item for item in all_news_data if datetime.strptime(item['date'], '%Y-%m-%d').date() > last_run_date.date()]
    
    if not all_news_data:
        logging.warning(f"No news data fetched for {stock_name}.")
        return

    save_path = os.path.join(settings['save_path'], f"news_data_{stock_code_padded}.csv")
    save_news(all_news_data, save_path)

# Function to fetch news for all stocks
def fetch_news_for_all_stocks(last_run_date):
    logging.info("Fetching news for all stocks...")
    stock_codes_and_names = get_stock_codes_and_names()
    if not stock_codes_and_names:
        logging.error("No stock codes and names found.")
        return

    with ThreadPoolExecutor(max_workers=10) as executor:
        futures = [executor.submit(update_news_data, stock['code'], stock['name'], last_run_date) for stock in stock_codes_and_names]
        for future in as_completed(futures):
            future.result()
    logging.info("Finished fetching news for all stocks.")

# Function to get corporate codes
def get_corp_codes(api_key):
    logging.info("Fetching corporate codes...")
    response = requests.get(corp_code_url, params={'crtfc_key': api_key})
    if response.status_code == 200:
        try:
            with open('corp_code.zip', 'wb') as f:
                f.write(response.content)
            with zipfile.ZipFile('corp_code.zip', 'r') as zip_ref:
                zip_ref.extractall()
            tree = ET.parse('CORPCODE.xml')
            root = tree.getroot()
            corp_codes = {}
            for child in root:
                stock_code = child.find('stock_code').text
                corp_code = child.find('corp_code').text
                if stock_code:
                    corp_codes[stock_code.zfill(6)] = corp_code  # 6자리로 맞추기
            return corp_codes
        except zipfile.BadZipFile:
            logging.error("Downloaded file is not a valid ZIP file.")
        except ET.ParseError:
            logging.error("Failed to parse XML from the ZIP file.")
        except Exception as e:
            logging.error(f"An unexpected error occurred: {e}")
    else:
        logging.error(f"Failed to download corporate codes. HTTP Status: {response.status_code}")
    return {}

corp_codes = get_corp_codes(dart_api_key)
logging.info(f"Loaded corporate codes: {list(corp_codes.keys())}")

# Function to get financial data
def get_financial_data(corp_code, stock_code, year, reprt_code='11011', fs_div='CFS'):
    logging.info(f"Fetching financial data for {stock_code} (corp code: {corp_code}, year: {year})...")
    try:
        params = {
            'crtfc_key': dart_api_key,
            'corp_code': corp_code,
            'bsns_year': str(year),
            'reprt_code': reprt_code,  # 사업보고서
            'fs_div': fs_div  # 연결재무제표
        }
        response = requests.get(fnltt_url, params=params)
        response.raise_for_status()
        data = response.json()
        if 'list' in data:
            df = pd.DataFrame(data['list'])
            df['date'] = f'{year}-12-31'  # Add date column with end of year
            return df
        else:
            logging.warning(f"No financial data found for {stock_code}. Response: {data}")
            return None
    except requests.exceptions.RequestException as e:
        logging.error(f"Failed to fetch financial data for {stock_code}. Error: {e}")
        return None

# Function to fetch and save financial data for all stocks
def fetch_and_save_financial_data(last_run_date):
    stock_count = 0  # Initialize a counter for tracking the number of stocks processed

    for stock in stock_codes_and_names:
        stock_code = stock['code'].zfill(6)  # 종목 코드를 6자리로 맞춤
        corp_code = corp_codes.get(stock_code)
        if corp_code:
            all_data = []
            for year in range(start_year, current_year + 1):
                # Skip if the data was already collected after the last run date
                if last_run_date and year <= last_run_date.year:
                    continue

                data = get_financial_data(corp_code, stock_code, year)
                if data is not None:
                    all_data.append(data)
                else:
                    logging.warning(f"No financial data for {stock_code} in year {year}")
            if all_data:
                combined_df = pd.concat(all_data, ignore_index=True)
                save_path = os.path.join(settings['financial_data_path'], f'{stock_code}_financial_statements.csv')
                combined_df.to_csv(save_path, index=False, encoding='utf-8-sig')
                logging.info(f"{stock_code} financial data saved to '{save_path}'")
        else:
            logging.warning(f"No corporate code found for stock code {stock_code}")

        # Increment the counter after processing each stock
        stock_count += 1

        # Pause for 3 minutes every 20 stocks
        if stock_count % 20 == 0:
            logging.info("Processed 20 stocks, taking a 3-minute break...")
            time.sleep(180)  # Sleep for 180 seconds (3 minutes)

# API 데이터 가져오기 함수 (XML 형식)
def fetch_data(api_key, stat_code, year, period_type="A"):
    url = f"https://ecos.bok.or.kr/api/StatisticSearch/{api_key}/xml/kr/1/100/{stat_code}/{period_type}/{year}/{year}"
    response = requests.get(url)
    if response.status_code == 200:
        return response.text
    else:
        raise Exception(f"API 호출 실패: {response.status_code}, {response.text}")

# 데이터 파싱 함수 (XML 형식)
def parse_data(xml_data, key):
    root = ET.fromstring(xml_data)
    items = root.findall('.//row')
    data = [{'date': item.find('TIME').text, 'Item Code': item.find('ITEM_CODE1').text, 'Item Name': item.find('ITEM_NAME1').text, key: item.find('DATA_VALUE').text} for item in items]
    return pd.DataFrame(data)

# 데이터 수집 함수
def collect_data(api_key, stat_code, indicator_name):
    collected_data = pd.DataFrame()
    for year in range(start_year, current_year + 1):
        xml_data = fetch_data(api_key, stat_code, year, "A")
        yearly_data = parse_data(xml_data, indicator_name)
        collected_data = pd.concat([collected_data, yearly_data], ignore_index=True)
    return collected_data

# 데이터 수집 및 저장 함수
def save_data(api_key, stat_code, indicator_name, save_path):
    data = collect_data(api_key, stat_code, indicator_name)
    data.to_csv(save_path, index=False)
    logging.info(f"{indicator_name} data saved to {save_path}")

# Main function to handle data collection
def main():
    logging.info("Starting data collection...")

    try:
        # Fetch the last run date
        last_run_date = get_last_run_date()

        # Define the current year and start year
        global current_year, start_year  # 전역 변수로 선언하여 다른 함수에서도 사용 가능
        current_year = datetime.now().year
        start_year = current_year - 10
        
        logging.info(f"Current year: {current_year}, Start year: {start_year}")

        # Fetch economic data if it's a new day (collect this data first)
        if last_run_date is None or last_run_date.date() < datetime.now().date():
            gdp_save_path = os.path.join(settings['economic_data_path'], 'gdp_data.csv')
            inflation_rate_save_path = os.path.join(settings['economic_data_path'], 'inflation_rate_data.csv')

            save_data(ecos_api_key, indicators["GDP"], "GDP", gdp_save_path)
            save_data(ecos_api_key, indicators["Inflation Rate"], "Inflation Rate", inflation_rate_save_path)

            # Update last run date
            save_last_run_date()

        # Initialize stock codes and names
        global stock_codes_and_names  # 전역 변수로 선언하여 다른 함수에서도 사용 가능
        stock_codes_and_names = get_stock_codes_and_names()
        
        # Fetch news data for stocks
        fetch_news_for_all_stocks(last_run_date)

        # Fetch financial data
        fetch_and_save_financial_data(last_run_date)

        # New stock detection
        old_stock_codes = get_stock_codes_and_names()
        new_stock_codes = fetch_and_save_stock_codes()
        new_codes = find_new_stock_codes(old_stock_codes, new_stock_codes)

        if new_codes:
            logging.info(f"New stock codes detected: {new_codes}")
            for new_stock in new_codes:
                stock_code = new_stock['code']
                stock_name = new_stock['name']
                update_news_data(stock_code, stock_name, last_run_date)

                corp_code = corp_codes.get(stock_code.zfill(6))
                if corp_code:
                    all_data = []
                    for year in range(start_year, current_year + 1):
                        if last_run_date and year <= last_run_date.year:
                            continue

                        data = get_financial_data(corp_code, stock_code, year)
                        if data is not None:
                            all_data.append(data)
                        else:
                            logging.warning(f"No financial data for {stock_code} in year {year}")
                    if all_data:
                        combined_df = pd.concat(all_data, ignore_index=True)
                        save_path = os.path.join(settings['financial_data_path'], f'{stock_code}_financial_statements.csv')
                        combined_df.to_csv(save_path, index=False, encoding='utf-8-sig')
                        logging.info(f"{stock_code} financial data saved to '{save_path}'")
                else:
                    logging.warning(f"No corporate code found for stock code {stock_code}")

    except Exception as e:
        logging.error(f"An error occurred: {e}. Retrying in 1 hour.")
        time.sleep(3600)
        main()



# Schedule tasks
schedule.every().day.at("08:00").do(main)  # 평일 기준 매일 아침 8시에 실행
# Initial run
main()

# Main loop
last_checked_minute = datetime.now().minute  # Track the last minute checked
while True:
    now = datetime.now()

    try:
        # Check for new stocks once every minute
        if now.minute != last_checked_minute:
            logging.info("Checking for new stock codes...")
            old_stock_codes = get_stock_codes_and_names()
            new_stock_codes = fetch_and_save_stock_codes()
            new_codes = find_new_stock_codes(old_stock_codes, new_stock_codes)

            if new_codes:
                logging.info(f"New stock codes detected: {new_codes}")
                for new_stock in new_codes:
                    stock_code = new_stock['code']
                    stock_name = new_stock['name']
                    update_news_data(stock_code, stock_name, last_run_date)

                    corp_code = corp_codes.get(stock_code.zfill(6))
                    if corp_code:
                        all_data = []
                        for year in range(start_year, current_year + 1):
                            if last_run_date and year <= last_run_date.year:
                                continue

                            data = get_financial_data(corp_code, stock_code, year)
                            if data is not None:
                                all_data.append(data)
                            else:
                                logging.warning(f"No financial data for {stock_code} in year {year}")
                        if all_data:
                            combined_df = pd.concat(all_data, ignore_index=True)
                            save_path = os.path.join(settings['financial_data_path'], f'{stock_code}_financial_statements.csv')
                            combined_df.to_csv(save_path, index=False, encoding='utf-8-sig')
                            logging.info(f"{stock_code} financial data saved to '{save_path}'")
                    else:
                        logging.warning(f"No corporate code found for stock code {stock_code}")

            last_checked_minute = now.minute

        # Run pending schedule tasks
        schedule.run_pending()

    except Exception as e:
        logging.error(f"An error occurred in the main loop: {e}. Retrying in 1 hour.")
        time.sleep(3600)

    time.sleep(1)
